:- module(analyzer_tab,[
	go_p/1,
	analyze/1
			]).


:- use_package(library(tabling)).
:- use_package(library(abs_solver)).


:- use_module(library(prolog_sys)).
:- use_module(library(write)).
:- use_module(library(format)).
:- use_module(library(terms_check)).
:- use_module(library(aggregates)).

:- multifile stored_program/2.


cputime(T) :-
  statistics(runtime,[T,_]).

writeln(T) :-
  write(T),
  nl.

go_p(Program) :- 
  cputime(Start),
      analyze(Program),
  cputime(End),
  T is End - Start,
  write('% test_tabling.pl: execution time = '), format('~2f',T), write(' milliseconds    '),nl, display(Program),nl.

:- table step_goal_table/3.

analyze(Name/L) :- !,
	length(List,L),
	inizialize(List),
	Goal =.. [Name | List],
	analyze(Goal),
	abolish_all_tables.

analyze(Goal) :-
	Goal =.. [_ | List],
	inizialize(List),
	(
	    findall(List,
	    (
		step_att(Goal),
		check_no_bot(List) 
	    ), GoalClauses),
	    join_info_att(GoalClauses, List) ->
	    true
	;
	    display('fail'),nl,
	    make_bot(List)
	),
	abolish_all_tables.

inizialize([]).
inizialize([X | Xs]) :-
	abs(X, top),
	inizialize(Xs).

dump_info(In, Out) :-
	get_abs_list(In, Out).

interp_att(Goal, InfoIn, InfoOut) :-
	Goal =..[_ | List],
	put_abs_list(List, InfoIn),
	(
	    step_att(Goal) ->
	    true
	;
	    make_bot(List)
	),
	get_abs_list(List, InfoOut),
	abolish_all_tables.

interp_body_att([]).
interp_body_att([SubGoal | RestBody]) :-
	step_att(SubGoal),
	interp_body_att(RestBody).

step_goal_att(Goal) :-
	Goal =.. [_ | List],
	get_abs_list(List, InfoIn),
	step_goal_table(Goal, InfoIn, InfoOut),
	put_abs_list(_, InfoOut).

step_goal_table(Goal, InfoIn, InfoOut) :-
	Goal =.. [_ | List],
	put_abs_list(_, InfoIn),
	stored_program(Goal, Body),
	interp_body_att(Body),
	get_abs_list(List, InfoOut).


step_att(Goal) :-
	stored_program(Goal,_), !,
	step_goal_att(Goal).


step_att(V is E) :- 	V abs_is E.

step_att(atom(V)) :- 	abs_atom(V).

step_att(V1 = V2) :- 	V1 .=. V2.

step_att(V1 < V2):- 	V1 .<. V2.
step_att(V1 > V2):- 	V1 .>. V2.
step_att(V1 =:= V2):- 	V1 .=:=. V2.

step_att(V1 @< V2):- 	V1 .@<. V2.
step_att(V1 @> V2):- 	V1 .@>. V2.
step_att(V1 == V2):- 	V1 .==. V2.


% abstract operators
step_att(V1 .<-. AbsV1)  :- 	V1 .<-. AbsV1.
step_att(abs(V1, AbsV1)) :- 	abs(V1, AbsV1).
step_att(V1 in AbsV1)    :-     ground(AbsV1), !, V1 .<-. AbsV1.
step_att(V1 in V2)       :- 	V1 in V2.



